#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Tests how the physical blockdag construction logic works.
This refers to how parent hashes and signatures are built.

Most of these tests will be asserting the obvious, with the exception of Reproducing behaviour.
"""

import unittest
import logging

from dlg.common.reproducibility.constants import (
    ReproducibilityFlags,
    ALL_RMODES,
)
from dlg.common.reproducibility.reproducibility import build_blockdag

logger = logging.getLogger("__name__")


def _generate_dummy_compute(rmode: ReproducibilityFlags):
    if rmode is not ReproducibilityFlags.ALL:
        return {
            "oid": 1,
            "reprodata": {
                "rmode": str(rmode.value),
                rmode.name: {
                    "lgt_data": {
                        "categoryType": "Application",
                        "category": "BashShellApp",
                    },
                    "lg_blockhash": "1",
                    "pgt_data": {"merkleroot": "2"},
                    "pgt_parenthashes": {},
                    "pgt_blockhash": "3",
                    "pg_data": {"merkleroot": "4"},
                    "pg_parenthashes": {},
                    "pg_blockhash": "5",
                    "rg_data": {"merkleroot": "6"},
                    "rg_parenthashes": {},
                },
            },
        }
    else:
        out_val = {"oid": 1, "reprodata": {}}
        out_val["reprodata"]["rmode"] = str(rmode.value)
        for level in ALL_RMODES:
            out_val["reprodata"][level.name] = {
                "lgt_data": {
                    "categoryType": "Application",
                    "category": "BashShellApp",
                },
                "lg_blockhash": "1",
                "pgt_data": {"merkleroot": "2"},
                "pgt_parenthashes": {},
                "pgt_blockhash": "3",
                "pg_data": {"merkleroot": "4"},
                "pg_parenthashes": {},
                "pg_blockhash": "5",
                "rg_data": {"merkleroot": "6"},
                "rg_parenthashes": {},
            }
        return out_val


def _generate_dummy_data(rmode: ReproducibilityFlags):
    if rmode is not ReproducibilityFlags.ALL:
        return {
            "oid": 1,
            "reprodata": {
                "rmode": str(rmode.value),
                rmode.name: {
                    "lgt_data": {
                        "categoryType": "Data",
                        "category": "File",
                    },
                    "lg_blockhash": "a",
                    "pgt_data": {"merkleroot": "b"},
                    "pgt_parenthashes": {},
                    "pgt_blockhash": "c",
                    "pg_data": {"merkleroot": "d"},
                    "pg_parenthashes": {},
                    "pg_blockhash": "e",
                    "rg_data": {"merkleroot": "f"},
                    "rg_parenthashes": {},
                },
            },
        }
    else:
        out_val = {"oid": 1, "reprodata": {}}
        out_val["reprodata"]["rmode"] = str(rmode.value)
        for level in ALL_RMODES:
            out_val["reprodata"][level.name] = {
                "lgt_data": {
                    "categoryType": "Data",
                    "category": "File",
                },
                "lg_blockhash": "a",
                "pgt_data": {"merkleroot": "b"},
                "pgt_parenthashes": {},
                "pgt_blockhash": "c",
                "pg_data": {"merkleroot": "d"},
                "pg_parenthashes": {},
                "pg_blockhash": "e",
                "rg_data": {"merkleroot": "f"},
                "rg_parenthashes": {},
            }
        return out_val


def _init_pgraph_single(rmode: ReproducibilityFlags):
    pgt = [_generate_dummy_compute(rmode)]
    return pgt


def _init_pgraph_twostart(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
    ]
    pgt[1]["oid"] = 2
    pgt[2]["oid"] = 3
    pgt[0]["outputs"] = [2]
    pgt[2]["outputs"] = [2]
    return pgt


def _init_pgraph_twoend(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
    ]
    pgt[1]["oid"] = 2
    pgt[2]["oid"] = 3
    pgt[0]["outputs"] = [2, 3]
    for drop in pgt:
        drop["reprodata"]["rmode"] = str(rmode.value)
    return pgt


def _init_pgraph_twolines(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_compute(rmode),
    ]
    pgt[1]["oid"] = 2
    pgt[2]["oid"] = 3
    pgt[3]["oid"] = 4
    pgt[0]["outputs"] = [2]
    pgt[2]["outputs"] = [4]
    for drop in pgt:
        drop["reprodata"]["rmode"] = str(rmode.value)
    return pgt


def _init_pgraph_data_fan(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_data(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_data(rmode),
        _generate_dummy_data(rmode),
    ]
    for i, drop in enumerate(pgt):
        pgt[i]["oid"] = i
        pgt[i]["reprodata"]["rmode"] = str(rmode.value)
    pgt[0]["outputs"] = [1]
    pgt[1]["outputs"] = [2, 3]
    return pgt


def _init_pgraph_data_funnel(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_data(rmode),
        _generate_dummy_data(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_data(rmode),
    ]
    for i, drop in enumerate(pgt):
        pgt[i]["oid"] = i
        pgt[i]["reprodata"]["rmode"] = str(rmode.value)
    pgt[0]["outputs"] = [2]
    pgt[1]["outputs"] = [2]
    pgt[2]["outputs"] = [3]
    return pgt


def _init_pgraph_data_sandwich(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_data(rmode),
        _generate_dummy_compute(rmode),
        _generate_dummy_data(rmode),
    ]
    for i, drop in enumerate(pgt):
        pgt[i]["oid"] = i
        pgt[i]["reprodata"]["rmode"] = str(rmode.value)
    pgt[0]["outputs"] = [1]
    pgt[1]["outputs"] = [2]
    return pgt


def _init_pgraph_computation_sandwich(rmode: ReproducibilityFlags):
    pgt = [
        _generate_dummy_compute(rmode),
        _generate_dummy_data(rmode),
        _generate_dummy_compute(rmode),
    ]
    for i, drop in enumerate(pgt):
        pgt[i]["oid"] = i
        pgt[i]["reprodata"]["rmode"] = str(rmode.value)
    pgt[0]["outputs"] = [1]
    pgt[1]["outputs"] = [2]
    return pgt


class PhysicalBlockdagRerunTests(unittest.TestCase):
    """
    Tests physical blockdag construction when rerunning.
    This should be relatively straightforward, but expanded group nodes scatter/gather etc. are
    special cases for rerunning specifically.
    """

    rmode = ReproducibilityFlags.RERUN

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            len(leaves) == 1
            and len(parenthashes) == 2
            and parenthashes[0] == parenthashes[1]
        )

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[2]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )


class PhysicalBlockdagRepeatTests(unittest.TestCase):
    """
    Tests physical blockdag construction when repeating.
    This should be relatively straightforward.
    """

    rmode = ReproducibilityFlags.REPEAT

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            len(leaves) == 1
            and len(parenthashes) == 2
            and parenthashes[0] == parenthashes[1]
        )

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[2]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )


class PhysicalBlockdagRecomputeTests(unittest.TestCase):
    """
    Tests physical blockdag construction when recomputing.
    This should be relatively straightforward.
    """

    rmode = ReproducibilityFlags.RECOMPUTE

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            len(leaves) == 1
            and len(parenthashes) == 2
            and parenthashes[0] == parenthashes[1]
        )

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[2]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )


class PhysicalBlockdagReproduceTests(unittest.TestCase):
    """
    Tests physical blockdag construction when reproducing.
    Rerunning should bring about contracting behaviour when data drops are introduced.
    """

    rmode = ReproducibilityFlags.REPRODUCE

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(len(leaves) == 1 and len(parenthashes) == 0)

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[0]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[0]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 2
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[0]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        # Not going to get anything out of this, since reproduce only cares about terminal data.
        self.assertEqual(0, len(parenthashes))


class PhysicalBlockdagReplicateScientificTests(unittest.TestCase):
    """
    Tests physical blockdag construction when replicating scientifically.
    This should be relatively straightforward.
    """

    rmode = ReproducibilityFlags.REPLICATE_SCI

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            len(leaves) == 1
            and len(parenthashes) == 2
            and parenthashes[0] == parenthashes[1]
        )

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[2]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )


class PhysicalBlockdagReplicateComputationTests(unittest.TestCase):
    """
    Tests physical blockdag construction when replicating computationally.
    This should be relatively straightforward.
    """

    rmode = ReproducibilityFlags.REPLICATE_COMP

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            len(leaves) == 1
            and len(parenthashes) == 2
            and parenthashes[0] == parenthashes[1]
        )

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[2]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )


class PhysicalBlockdagReplicateTotalTests(unittest.TestCase):
    """
    Tests physical blockdag construction when replicating totally.
    This should be relatively straightforward.
    """

    rmode = ReproducibilityFlags.REPLICATE_TOTAL

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        parenthashes = list(
            pgr[1]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            len(leaves) == 1
            and len(parenthashes) == 2
            and parenthashes[0] == parenthashes[1]
        )

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        leaves = build_blockdag(pgr, "pg", self.rmode)[0]
        self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthash1 = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        parenthash2 = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            parenthash1 == parenthash2 and parenthash1[0] == sourcehash
        )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[2]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[3]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        build_blockdag(pgr, "pg", self.rmode)
        sourcehash = pgr[1]["reprodata"][self.rmode.name]["pg_blockhash"]
        parenthashes = list(
            pgr[2]["reprodata"][self.rmode.name]["pg_parenthashes"].values()
        )
        self.assertTrue(
            sourcehash == parenthashes[0] and len(parenthashes) == 1
        )


class PhysicalBlockdagAllTests(unittest.TestCase):
    """
    Tests physical blockdag construction when test ALL mode.
    This should be relatively straightforward, but expanded group nodes scatter/gather etc. are
    special cases for rerunning specifically.
    """

    rmode = ReproducibilityFlags.ALL

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single(self.rmode)
        for rmode in ALL_RMODES:
            leaves = build_blockdag(pgr, "pg", rmode)[0]
            self.assertTrue(len(leaves) == 1)

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart(self.rmode)
        for rmode in ALL_RMODES:
            leaves = build_blockdag(pgr, "pg", rmode)[0]
            parenthashes = list(
                pgr[1]["reprodata"][rmode.name]["pg_parenthashes"].values()
            )
            if rmode is not ReproducibilityFlags.REPRODUCE:
                self.assertTrue(
                    len(leaves) == 1
                    and len(parenthashes) == 2
                    and parenthashes[0] == parenthashes[1]
                )
            else:
                self.assertTrue(len(leaves) == 1 and len(parenthashes) == 0)

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend(self.rmode)
        for rmode in ALL_RMODES:
            leaves = build_blockdag(pgr, "pg", rmode)[0]
            self.assertTrue(leaves[0] == leaves[1])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines(self.rmode)
        for rmode in ALL_RMODES:
            leaves = build_blockdag(pgr, "pg", rmode)[0]
            self.assertTrue(leaves[0] == leaves[1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        pgr = _init_pgraph_data_fan(self.rmode)
        for rmode in ALL_RMODES:
            build_blockdag(pgr, "pg", rmode)
            if rmode is ReproducibilityFlags.REPRODUCE:
                sourcehash = pgr[0]["reprodata"][rmode.name]["pg_blockhash"]
            else:
                sourcehash = pgr[1]["reprodata"][rmode.name]["pg_blockhash"]
            parenthash1 = list(
                pgr[2]["reprodata"][rmode.name]["pg_parenthashes"].values()
            )
            parenthash2 = list(
                pgr[3]["reprodata"][rmode.name]["pg_parenthashes"].values()
            )
            self.assertTrue(
                parenthash1 == parenthash2 and parenthash1[0] == sourcehash
            )

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        pgr = _init_pgraph_data_funnel(self.rmode)
        for rmode in ALL_RMODES:
            build_blockdag(pgr, "pg", rmode)
            if rmode is ReproducibilityFlags.REPRODUCE:
                sourcehash = pgr[0]["reprodata"][rmode.name]["pg_blockhash"]
                parenthashes = list(
                    pgr[3]["reprodata"][rmode.name]["pg_parenthashes"].values()
                )
                self.assertTrue(
                    sourcehash == parenthashes[0] and len(parenthashes) == 2
                )
            else:
                sourcehash = pgr[2]["reprodata"][rmode.name]["pg_blockhash"]
                parenthashes = list(
                    pgr[3]["reprodata"][rmode.name]["pg_parenthashes"].values()
                )
                self.assertTrue(
                    sourcehash == parenthashes[0] and len(parenthashes) == 1
                )

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        pgr = _init_pgraph_data_sandwich(self.rmode)
        for rmode in ALL_RMODES:
            build_blockdag(pgr, "pg", rmode)
            if rmode is ReproducibilityFlags.REPRODUCE:
                sourcehash = pgr[0]["reprodata"][rmode.name]["pg_blockhash"]
            else:
                sourcehash = pgr[1]["reprodata"][rmode.name]["pg_blockhash"]
            parenthashes = list(
                pgr[2]["reprodata"][rmode.name]["pg_parenthashes"].values()
            )
            self.assertTrue(
                sourcehash == parenthashes[0] and len(parenthashes) == 1
            )

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        pgr = _init_pgraph_computation_sandwich(self.rmode)
        for rmode in ALL_RMODES:
            build_blockdag(pgr, "pg", rmode)
            sourcehash = pgr[1]["reprodata"][rmode.name]["pg_blockhash"]
            parenthashes = list(
                pgr[2]["reprodata"][rmode.name]["pg_parenthashes"].values()
            )
            if rmode != ReproducibilityFlags.REPRODUCE:
                self.assertTrue(
                    sourcehash == parenthashes[0] and len(parenthashes) == 1
                )
            else:
                self.assertTrue(len(parenthashes) == 0)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import base64
import io
import logging
import os

from benedict import benedict

from dlg.data.drops.data_base import DataDROP
from dlg.drop import DEFAULT_INTERNAL_PARAMETERS
from dlg.data.io import MemoryIO
from dlg.meta import dlg_string_param

logger = logging.getLogger(__name__)


##
# @brief ParameterSet
# @details A set of parameters, which can be set and modified in EAGLE and thus is part of the graph. Multiple serialisation formats are available.
# @par EAGLE_START
# @param category ParameterSet
# @param tag daliuge
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param mode "YANDA"/Select/ComponentParameter/NoPort/ReadWrite/YANDA,ini,yaml,json,toml,pickle/False/False/Serialisation method.
# @param config_data ""/String/ComponentParameter/NoPort/ReadWrite//False/False/Additional configuration information to be mixed in with the initial data
# @param dropclass dlg.data.drops.parset_drop.ParameterSetDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param Config /Object.File/ApplicationArgument/OutputPort/ReadWrite//False/False/The output configuration file
# @par EAGLE_END
class ParameterSetDROP(DataDROP):
    """
    A generic configuration drop template wrapper
    This drop opens an (optional) file containing some initial configuration information, then
    appends any additional specified parameters to it, finally serving it as a data object.
    """

    config_data = b""

    mode = dlg_string_param("mode", None)

    def serialize_parameters(self, parameters: dict, mode):
        """
        Returns a string representing a serialization of the parameters.
        """
        if mode not in ["ini", "yaml", "json", "toml", "YANDA", "pickle"]:
            mode = "yaml"
        if mode == "YANDA":
            # TODO: Add more complex value checking
            serialp = "\n".join(f"{x}={y.value}" for x, y in parameters.items())
            serialp = serialp.encode("utf-8")
        # Add more formats (.ini for example)
        elif mode == "ini":
            serialp = parameters.to_ini().encode("utf-8")
        elif mode == "yaml":
            serialp = parameters.to_yaml().encode("utf-8")
        elif mode == "json":
            serialp = parameters.to_json().encode("utf-8")
        elif mode == "toml":
            serialp = parameters.to_toml().encode("utf-8")
        elif mode == "pickle":
            serialp = base64.b64decode(parameters.to_pickle())
        return serialp

    def filter_parameters(self, parameters: dict, mode):
        """
        Returns a dictionary of parameters, with daliuge-internal or other parameters filtered out
        """
        if mode == "YANDA_DEPRECATED":  # we now return all applicationArgs
            forbidden_params = list(DEFAULT_INTERNAL_PARAMETERS)
            if parameters["config_data"] == b"":
                forbidden_params.append("configData")
            return {
                key: val
                for key, val in parameters.items()
                if key not in DEFAULT_INTERNAL_PARAMETERS
            }
        # we use benedict to get access to the serializers.
        return benedict(
            {
                k: benedict(
                    {
                        "value": v.value,
                        "description": v.description,
                        "type": v.type,
                        "default": v.defaultValue,
                    },
                    keypath_separator=None,
                )
                for k, v in benedict(
                    parameters["applicationArgs"], keypath_separator=None
                ).items()
            },
            keypath_separator=None,  # support "." in keyword name
        )

    def initialize(self, **kwargs):
        """
        TODO: Open input file
        """
        logger.debug(
            "Application arguments found: %s", self.parameters["applicationArgs"]
        )
        self.config_data = self.serialize_parameters(
            self.filter_parameters(self.parameters, self.mode), self.mode
        )

    def getIO(self):
        return MemoryIO(io.BytesIO(self.config_data))

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]
        return f"config://{hostname}/{os.getpid()}/{id(self.config_data)}"



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import contextlib
import importlib

from dlg.data.drops.data_base import DataDROP, logger
from dlg.exceptions import InvalidDropException
from dlg.data.io import ErrorIO
from dlg.meta import dlg_dict_param
from dlg.utils import prepare_sql


##
# @brief RDBMS
# @details A Drop allowing storage and retrieval from a SQL DB.
# @par EAGLE_START
# @param category RDBMS
# @param tag daliuge
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param dbmodule /String/ComponentParameter/NoPort/ReadWrite//False/False/Load path for python DB module
# @param dbtable /String/ComponentParameter/NoPort/ReadWrite//False/False/The name of the table to use
# @param vals {}/Json/ComponentParameter/NoPort/ReadWrite//False/False/Json encoded values dictionary used for INSERT. The keys of ``vals`` are used as the column names.
# @param condition /String/ComponentParameter/NoPort/ReadWrite//False/False/Condition for SELECT. For this the WHERE statement must be written using the "{X}" or "{}" placeholders
# @param selectVals {}/Json/ComponentParameter/NoPort/ReadWrite//False/False/Values for the WHERE statement
# @param dropclass dlg.data.drops.rdbms.RDBMSDrop/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param dummy /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Dummy port
# @par EAGLE_END
class RDBMSDrop(DataDROP):
    """
    A Drop that stores data in a table of a relational database
    """

    dbparams = dlg_dict_param("dbparams", {})

    def initialize(self, **kwargs):
        DataDROP.initialize(self, **kwargs)

        if "dbmodule" not in kwargs:
            raise InvalidDropException(
                self, '%r needs a "dbmodule" parameter' % (self,)
            )
        if "dbtable" not in kwargs:
            raise InvalidDropException(
                self, '%r needs a "dbtable" parameter' % (self,)
            )

        # The DB-API 2.0 module
        dbmodname = kwargs.pop("dbmodule")
        self._db_drv = importlib.import_module(dbmodname)

        # The table this Drop points at
        self._db_table = kwargs.pop("dbtable")

        # Data store for reproducibility
        self._querylog = []

    def getIO(self):
        # This Drop cannot be accessed directly
        return ErrorIO()

    def _connection(self):
        return contextlib.closing(self._db_drv.connect(**self.dbparams))

    def _cursor(self, conn):
        return contextlib.closing(conn.cursor())

    def insert(self, vals: dict):
        """
        Inserts the values contained in the ``vals`` dictionary into the
        underlying table. The keys of ``vals`` are used as the column names.
        """
        with self._connection() as c:
            with self._cursor(c) as cur:
                # vals is a dictionary, its keys are the column names and its
                # values are the values to insert
                sql = "INSERT into %s (%s) VALUES (%s)" % (
                    self._db_table,
                    ",".join(vals.keys()),
                    ",".join(["{}"] * len(vals)),
                )
                sql, vals = prepare_sql(
                    sql, self._db_drv.paramstyle, list(vals.values())
                )
                logger.debug(
                    "Executing SQL with parameters: %s / %r", sql, vals
                )
                cur.execute(sql, vals)
                c.commit()

    def select(self, columns=None, condition=None, vals=()):
        """
        Returns the selected values from the table. Users can constrain the
        result set by specifying a list of ``columns`` to be returned (otherwise
        all table columns are returned) and a ``condition`` to be applied,
        in which case a list of ``vals`` to be applied as query parameters can
        also be given.
        """
        with self._connection() as c:
            with self._cursor(c) as cur:
                # Build up SQL with optional columns and conditions
                columns = columns or ("*",)
                sql = [
                    "SELECT %s FROM %s" % (",".join(columns), self._db_table)
                ]
                if condition:
                    sql.append(" WHERE ")
                    sql.append(condition)

                # Go, go, go!
                sql, vals = prepare_sql(
                    "".join(sql), self._db_drv.paramstyle, vals
                )
                logger.debug(
                    "Executing SQL with parameters: %s / %r", sql, vals
                )
                cur.execute(sql, vals)
                if cur.description:
                    ret = cur.fetchall()
                else:
                    ret = []
                self._querylog.append((sql, vals, ret))
                return ret

    @property
    def dataURL(self) -> str:
        return "rdbms://%s/%s/%r" % (
            self._db_drv.__name__,
            self._db_table,
            self._db_params,
        )

    # Override
    def generate_reproduce_data(self):
        return {"query_log": self._querylog}



"""
Split the measurementset using MSTransform along the frequency axis
Used as a template to be wrapped by the DROP framework

chen.wu@icrar.org
"""

import datetime
import logging
import os
import re
import subprocess
import sys
from optparse import OptionParser
from string import Template

logger = logging.getLogger(__name__)

ms_transform_tpl = """
mstransform(vis='${infile}',
            outputvis='${outfile}',
            regridms=T,
            restfreq='1420.405752MHz',
            mode='frequency',
            nchan=${no_chan},
            outframe='lsrk',
            interpolation='linear',
            veltype='radio',
            start='${freq1}MHz',
            width='${width_freq}kHz',
            spw='${spw_range}',
            combinespws=True,
            nspw=1,
            createmms=False,
            datacolumn="data")
"""


def freq_map(low_req, hi_req, *args):
    """
    Ported from Chiles AWS code by R.Dodson (richard.dodson@icrar.org)

    Return the Spectral Window Required given the lower and upper bounds requested.
    The requests are compared to the 3 possible freq ranges (starting at 941, 946 or 951 MHz).
    2 MHz buffer is added to these values.
    2MHz is equiv to 40km/s, which covers the range of calculated observatory velocities of ~+-30km/s

    Future: We could use the _actual_ spectral windows for each day to return a smaller range.
    Additional parameters would need to be passed (just the data file name?)

    Call doctest with # python -m doctest -v freq_map.py

    >>> freq_map(951,956)
    '0~0'

    >>> freq_map(951,983)
    '0~1'

    >>> freq_map(973,978,951)
    '0~0'

    >>> freq_map(973,983,951)
    '0~1'

    >>> freq_map(1005,1007,951)
    '1~1'

    >>> freq_map(1003,1007,941)
    '1~2'

    >>> freq_map(1003,1007,946)
    '1~1'

    >>> freq_map(1003,1007,951)
    '1~1'

    >>> freq_map(1007,1011,941)
    '2~2'

    >>> freq_map(1009,1013,946)
    '1~2'

    >>> freq_map(1009,1013,951)
    '1~2'

    >>> freq_map(1200,1210)
    '8~8'

    >>> freq_map(1360,1400)
    '13~14'

    >>> freq_map(1400,1404)
    '14~14'
    """

    #     SpwID  Name           #Chans   Frame   Ch0(MHz)   Ch0(MHz)   Ch0(MHz)ChanWid(kHz)  TotBW(kHz) BBC Num  Corrs
    #     0      EVLA_L#A0C0#0    2048   TOPO     941.000    946.000    951.000      15.625     32000.0      12  RR  LL
    #     1      EVLA_L#A0C0#1    2048   TOPO     973.000    978.000    983.000      15.625     32000.0      12  RR  LL
    #     2      EVLA_L#A0C0#2    2048   TOPO    1005.000   1010.000   1015.000      15.625     32000.0      12  RR  LL
    #     3      EVLA_L#A0C0#3    2048   TOPO    1037.000   1042.000   1047.000      15.625     32000.0      12  RR  LL
    #     4      EVLA_L#A0C0#4    2048   TOPO    1069.000   1074.000   1079.000      15.625     32000.0      12  RR  LL
    #     5      EVLA_L#A0C0#5    2048   TOPO    1101.000   1106.000   1111.000      15.625     32000.0      12  RR  LL
    #     6      EVLA_L#A0C0#6    2048   TOPO    1133.000   1138.000   1143.000      15.625     32000.0      12  RR  LL
    #     7      EVLA_L#A0C0#7    2048   TOPO    1165.000   1170.000   1175.000      15.625     32000.0      12  RR  LL
    #     8      EVLA_L#A0C0#8    2048   TOPO    1197.000   1202.000   1207.000      15.625     32000.0      12  RR  LL
    #     9      EVLA_L#A0C0#9    2048   TOPO    1229.000   1234.000   1239.000      15.625     32000.0      12  RR  LL
    #     10     EVLA_L#A0C0#10   2048   TOPO    1261.000   1266.000   1271.000      15.625     32000.0      12  RR  LL
    #     11     EVLA_L#A0C0#11   2048   TOPO    1293.000   1298.000   1303.000      15.625     32000.0      12  RR  LL
    #     12     EVLA_L#A0C0#12   2048   TOPO    1325.000   1330.000   1335.000      15.625     32000.0      12  RR  LL
    #     13     EVLA_L#A0C0#13   2048   TOPO    1357.000   1362.000   1367.000      15.625     32000.0      12  RR  LL
    #     14     EVLA_L#A0C0#14   2048   TOPO    1389.000   1394.000   1399.000      15.625     32000.0      12  RR  LL

    f_tab = [
        [941.00, 946.00, 951.00],
        [973.00, 978.00, 983.00],
        [1005.00, 1010.00, 1015.00],
        [1037.00, 1042.00, 1047.00],
        [1069.00, 1074.00, 1079.00],
        [1101.00, 1106.00, 1111.00],
        [1133.00, 1138.00, 1143.00],
        [1165.00, 1170.00, 1175.00],
        [1197.00, 1202.00, 1207.00],
        [1229.00, 1234.00, 1239.00],
        [1261.00, 1266.00, 1271.00],
        [1293.00, 1298.00, 1303.00],
        [1325.00, 1330.00, 1335.00],
        [1357.00, 1362.00, 1367.00],
        [1389.00, 1394.00, 1399.00],
        [1421.00, 1426.00, 1431.00],
    ]

    if_low = 0
    if_hi = 14
    ifn_low = 0
    ifn_hi = 2

    if args:
        if args[0] == 941:
            ifn_low = 0
            ifn_hi = 0
        elif args[0] == 946:
            ifn_low = 1
            ifn_hi = 1
        elif args[0] == 951:
            ifn_low = 2
            ifn_hi = 2

    for nif_low in range(0, 16):
        f = f_tab[nif_low]
        # print nif_low,f
        if (f[ifn_low]) > ((low_req) - 2):
            if_low = nif_low - 1
            # print 'Using '+str(if_low)+' for lower SPW edge'
            nif_low = 14
            break

    for nif_hi in range(0, 16):
        f = f_tab[nif_hi]
        # print nif_hi,f
        if f[ifn_hi] > (hi_req + 2):
            if_hi = nif_hi - 1
            # print 'Using '+str(if_hi)+' for upper SPW edge'
            nif_hi = 14
            break

    if if_low == -1:
        logger.info("Lower bound (%f) out of range", low_req)

    if if_hi == -1:
        logger.info("Upper bound (%f) out of range", hi_req)

    spw = "{0}~{1}".format(if_low, if_hi)
    return spw


def launch_mstransform(
    infile, outfile, no_chan, freq1, width_freq, spw_range, gen_script_dir, casa_bin_dir
):
    """
    launch CASA MSTransform as a separate (CASA) process

    Return: process reference
    """
    # each casapy must launch from a different directory
    mst_dir = "{0}/{1}".format(gen_script_dir, freq1)
    os.system("mkdir -p {0}".format(mst_dir))
    os.chdir(mst_dir)

    mst_dict = {}
    mst_dict["infile"] = infile
    mst_dict["outfile"] = outfile
    mst_dict["no_chan"] = no_chan
    mst_dict["freq1"] = freq1
    mst_dict["width_freq"] = width_freq
    mst_dict["spw_range"] = spw_range

    ts = Template(ms_transform_tpl)
    casa_script = ts.safe_substitute(mst_dict)

    gen_script_fn = "{0}/mstransform.py".format(mst_dir)
    with open(gen_script_fn, "w") as casa_file:
        casa_file.write(casa_script)

    casapy_cmd = "{0}/casapy --nologger -c {1}".format(casa_bin_dir, gen_script_fn)
    casa_process = subprocess.Popen(casapy_cmd.split())
    return casa_process


def do_split(
    infile, outdir, min_freq, max_freq, step_freq, width_freq, work_dir, casa_bin_dir
):
    """
    do actual mstransform, if outdir exists, it will be removed first

    TODO - add timer to measure completion time
    """

    dt = datetime.datetime.now()
    timestr = dt.strftime("%Y-%m-%dT%H-%M-%S")

    outdir += "/{0}".format(timestr)
    if os.path.exists(outdir):
        os.system("rm -rf {0}".format(outdir))
    os.system("mkdir -p {0}".format(outdir))

    steps = (max_freq - min_freq) / step_freq
    rem = (max_freq - min_freq) % step_freq
    if rem:
        steps += 1
    freq1 = min_freq
    freq2 = min_freq + step_freq
    bottom_edge = re.search("_[0-9]{3}_", infile)
    if bottom_edge:
        # e.g. 20131122_941_6_FINAL_PRODUCTS --> 941
        bedge = bottom_edge.group(0)
        bedge = int(bedge[1:4])
    else:
        bedge = None

    casa_proc_list = []
    gen_script_dir = "{0}/{1}".format(work_dir, timestr)

    for i in range(steps):  # potentially parallel
        spw_range = freq_map(freq1, freq2, bedge)
        if rem and (i == steps - 1):
            freq_range = "{0}~{1}".format(min_freq + i * step_freq, max_freq)
        else:
            freq_range = "{0}~{1}".format(freq1, freq2)

        no_chan = int(step_freq * 1000.0 / width_freq)  # MHz/kHz!!
        outfile = outdir + "vis_" + freq_range
        outfile = "{0}/vis_{1}".format(outdir, freq_range)
        logger.info("working on: %s", outfile)

        casa_proc = launch_mstransform(
            infile,
            outfile,
            no_chan,
            freq1,
            width_freq,
            spw_range,
            gen_script_dir,
            casa_bin_dir,
        )
        casa_proc_list.append(casa_proc)
        freq1 += step_freq
        freq2 += step_freq

    for csp in casa_proc_list:
        csp.wait()  # join all sub-processes before exiting


if __name__ == "__main__":
    """
    Setup all the parameters
    """
    parser = OptionParser()
    # compulsory directory parameters
    parser.add_option(
        "-i",
        "--input_ms",
        action="store",
        type="string",
        dest="infile",
        help="Input measurementset (Required)",
    )
    parser.add_option(
        "-o",
        "--output_dir",
        action="store",
        type="string",
        dest="outdir",
        help="Output directory (Required)",
    )
    parser.add_option(
        "-k",
        "--work_dir",
        action="store",
        type="string",
        dest="work_dir",
        help="Working directory (Required)",
    )
    parser.add_option(
        "-c",
        "--casa_bin_dir",
        action="store",
        type="string",
        dest="casa_bin_dir",
        help="CASA executable directory (Required)",
    )

    # optional mstransform parameters
    parser.add_option(
        "-n",
        "--min_freq",
        action="store",
        type="int",
        dest="min_freq",
        default=1020,
        help="Mininum frquency (MHz)",
    )
    parser.add_option(
        "-x",
        "--max_freq",
        action="store",
        type="int",
        dest="max_freq",
        default=1028,
        help="Maximum frquency (MHz)",
    )
    parser.add_option(
        "-s",
        "--freq_step",
        action="store",
        type="float",
        dest="step_freq",
        default=4,
        help="Subband width(MHz)",
    )
    parser.add_option(
        "-w",
        "--freq_width",
        action="store",
        type="float",
        dest="width_freq",
        default=15.625,
        help="Frquency resolution(KHz)",
    )

    (options, args) = parser.parse_args()
    if (
        None == options.infile
        or None == options.outdir
        or None == options.work_dir
        or None == options.casa_bin_dir
    ):
        parser.print_help()
        sys.exit(1)

    do_split(
        options.infile,
        options.outdir,
        options.min_freq,
        options.max_freq,
        options.step_freq,
        options.width_freq,
        options.work_dir,
        options.casa_bin_dir,
    )



import logging
import collections
from typing import Tuple
import dlg.common as common

logger = logging.getLogger(__name__)


def serialize_kwargs(keyargs, prefix="--", separator=" "):
    kwargs = []
    for name, value in iter(keyargs.items()):
        if prefix == "--" and len(name) == 1:
            kwargs += [f"-{name} {value}"]
        else:
            kwargs += [
                f"{prefix.strip()}{name.strip()}{separator}{str(value).strip()}"
            ]
    logger.debug("kwargs after serialization: %s", kwargs)
    return kwargs


def clean_applicationArgs(applicationArgs: dict) -> dict:
    """
    Removes arguments with None and False values, if not precious. This
    is in particular used for Bash and Docker app command lines, else
    we would have empty values for command line arguments.

    Args:
        applicationsArgs (dict): the complete set of arguments

    Returns:
        dict: a dictionary with the relevant arguments only.
    """
    cleanedArgs = {}
    if not isinstance(applicationArgs, dict):
        logger.info("applicationArgs are not passed as a dict. Ignored!")
    for name, vdict in applicationArgs.items():
        if vdict in [None, False, ""]:
            continue
        elif isinstance(vdict, bool):
            vdict = {"precious": False, "value": "", "positional": False}
        elif isinstance(vdict, dict):
            precious = vdict["precious"]
            if vdict["value"] in [None, False, ""] and not precious:
                continue
        cleanedArgs.update({name: vdict})
    logger.debug("After clean_applicationArgs: %s", cleanedArgs)
    return cleanedArgs


def serialize_applicationArgs(applicationArgs, prefix="--", separator=" "):
    """
    Unpacks the applicationArgs dictionary and returns two strings, one for
    positional arguments and one for kw arguments that can be used to construct
    the final command line.
    """
    applicationArgs = clean_applicationArgs(
        applicationArgs, prefix=prefix, separator=separator
    )
    pargs = []
    kwargs = {}
    for name, vdict in applicationArgs.items():
        value = vdict["value"]
        positional = vdict["positional"]
        if positional:
            pargs.append(str(value).strip())
        else:
            kwargs.update({name: value})
    skwargs = serialize_kwargs(kwargs, prefix=prefix, separator=separator)
    logger.info("Constructed command line arguments: %s %s", pargs, kwargs)
    return (pargs, skwargs)


def identify_named_ports(
    port_dict: dict,
    posargs: list,
    pargsDict: dict,
    keyargs: dict,
    check_len: int = 0,
    mode: str = "inputs",
) -> dict:
    """
    Checks port names for matches with arguments and returns mapped ports.

    Args:
        port_dict (dict): ports {uid:name,...}
        posargs (list): available positional arguments (will be modified)
        pargsDict (dict): mapped arguments (will be modified)
        keyargs (dict): keyword arguments
        check_len (int): number of of ports to be checked
        mode (str ["inputs"]): mode, used just for logging messages

    Returns:
        dict: port arguments

    Side effect:
        modifies pargsDict
    """
    # p_name = [p["name"] for p in port_dict]
    logger.debug(
        "Using named ports to remove %s from arguments port_dict: %s, check_len: %d)",
        mode,
        port_dict,
        check_len,
    )
    logger.debug("Checking against keyargs: %s", keyargs)
    portargs = {}
    posargs = list(posargs)
    keys = list(port_dict.keys())
    logger.debug("Checking ports: %s", keys)
    for i in range(check_len):
        try:
            key = port_dict[keys[i]]["name"]
            value = port_dict[keys[i]]["path"]
        except KeyError:
            logger.debug("portDict: %s", port_dict)
            raise KeyError
        if value is None:
            value = ""  # make sure we are passing NULL drop events
        if key in posargs:
            pargsDict.update({key: value})
            # portargs.update({key: value})
            logger.debug("Using %s '%s' for parg %s", mode, value, key)
            posargs.pop(posargs.index(key))
        elif key in keyargs:
            # if not found in appArgs we don't put them into portargs either
            portargs.update({key: value})
            # pargsDict.update({key: value})
            logger.debug(
                "Using %s of type %s for kwarg %s", mode, type(value), key
            )
            _ = keyargs.pop(key)  # remove from original arg list
        else:
            logger.debug(
                "No matching argument found for %s key %s, %s, %s",
                mode,
                key,
                keyargs,
                posargs,
            )
    logger.debug("Returning kw mapped ports: %s", portargs)
    return portargs


def check_ports_dict(ports: list) -> bool:
    """
    Checks whether all ports in ports list are of type dict. This is
    for backwards compatibility.

    Args:
        ports (list):

    Returns:
        bool: True if all ports are dict, else False
    """
    # all returns true if list is empty!
    if len(ports) > 0:
        return all(isinstance(p, dict) for p in ports)
    else:
        return False


def replace_named_ports(
    iitems: dict,
    oitems: dict,
    inport_names: dict,
    outport_names: dict,
    appArgs: dict,
    argumentPrefix: str = "--",
    separator: str = " ",
) -> Tuple[str, str]:
    """
    Function attempts to identify CLI component arguments that match port names.

    Inputs:
        iitems: itemized input port dictionary
        oitems: itemized output port dictionary
        inport_names: dictionary of input port names (key: uid)
        outport_names: dictionary of output port names (key: uid)
        appArgs: dictionary of all arguments
        argumentPrefix: prefix for keyword arguments
        separator: character used between keyword and value

    Returns:
        tuple of serialized keyword arguments and positional arguments
    """
    logger.debug(
        "iitems: %s; inport_names: %s; outport_names: %s",
        iitems,
        inport_names,
        outport_names,
    )
    inputs_dict = collections.OrderedDict()
    for uid, drop in iitems:
        inputs_dict[uid] = {"path": drop.path if hasattr(drop, "path") else ""}

    outputs_dict = collections.OrderedDict()
    for uid, drop in oitems:
        outputs_dict[uid] = {
            "path": drop.path if hasattr(drop, "path") else ""
        }
    # logger.debug("appArgs: %s", appArgs)
    # get positional args
    posargs = [arg for arg in appArgs if appArgs[arg]["positional"]]
    # get kwargs
    keyargs = {
        arg: appArgs[arg]["value"]
        for arg in appArgs
        if not appArgs[arg]["positional"]
    }
    # we will need an ordered dict for all positional arguments
    # thus we create it here and fill it with values
    portPosargsDict = collections.OrderedDict(
        zip(posargs, [None] * len(posargs))
    )
    logger.debug(
        "posargs: %s; keyargs: %s, %s",
        posargs,
        keyargs,
        check_ports_dict(inport_names),
    )
    portkeyargs = {}
    ipkeyargs = {}
    opkeyargs = {}
    if check_ports_dict(inport_names):
        for inport in inport_names:
            key = list(inport.keys())[0]
            inputs_dict[key].update({"name": inport[key]})

        ipkeyargs = identify_named_ports(
            inputs_dict,
            posargs,
            portPosargsDict,
            keyargs,
            check_len=len(iitems),
            mode="inputs",
        )
        portkeyargs.update(ipkeyargs)
    else:
        for i in range(min(len(iitems), len(posargs))):
            portkeyargs.update({list(posargs)[i]: list(iitems)[i][1]})

    if check_ports_dict(outport_names):
        for outport in outport_names:
            key = list(outport.keys())[0]
            outputs_dict[key].update({"name": outport[key]})
        opkeyargs = identify_named_ports(
            outputs_dict,
            posargs,
            portPosargsDict,
            keyargs,
            check_len=len(oitems),
            mode="outputs",
        )
        portkeyargs.update(opkeyargs)
    else:
        for i in range(min(len(oitems), len(posargs))):
            portkeyargs.update({posargs[i]: list(oitems)[i][1]})
    # now that we have the mapped ports we can cleanup the appArgs
    # and construct the final keyargs and pargs
    logger.debug(
        "Arguments from ports: %s, %s, %s, %s",
        portkeyargs,
        portPosargsDict,
        ipkeyargs,
        opkeyargs,
    )
    appArgs = clean_applicationArgs(appArgs)
    # get cleaned positional args
    posargs = {
        arg: appArgs[arg]["value"]
        for arg in appArgs
        if appArgs[arg]["positional"]
    }
    logger.debug("posargs: %s", posargs)
    # get cleaned kwargs
    keyargs = {
        arg: appArgs[arg]["value"]
        for arg in appArgs
        if not appArgs[arg]["positional"]
    }
    for k, v in portkeyargs.items():
        if v not in [None, ""]:
            keyargs.update({k: v})
    for k, v in portPosargsDict.items():
        logger.debug("port posarg %s has value %s", k, v)
        # logger.debug("default posarg %s has value %s", k, posargs[k])
        if k == "input_redirection":
            v = f"cat {v} > "
        if k == "output_redirection":
            v = f"> {v}"
        if v not in [None, ""]:
            posargs.update({k: v})
    keyargs = (
        serialize_kwargs(keyargs, prefix=argumentPrefix, separator=separator)
        if len(keyargs) > 0
        else [""]
    )
    pargs = list(posargs.values())
    pargs = [""] if len(pargs) == 0 or None in pargs else pargs
    logger.debug(
        "After port replacement: pargs: %s; keyargs: %s", pargs, keyargs
    )
    return keyargs, pargs



import heapq
import logging
from .data_base import DataDROP
from dlg.exceptions import InvalidRelationshipException
from sqlite3 import OperationalError
from dlg.data.io import (
    ErrorIO,
)
from dlg.ddap_protocol import (
    DROPLinkType,
    DROPRel,
)

logger = logging.getLogger(__name__)


class ContainerDROP(DataDROP):
    """
    A DROP that doesn't directly point to some piece of data, but instead
    holds references to other DROPs (its children), and from them its own
    internal state is deduced.

    Because of its nature, ContainerDROPs cannot be written to directly,
    and likewise they cannot be read from directly. One instead has to pay
    attention to its "children" DROPs if I/O must be performed.
    """

    def initialize(self, **kwargs):
        super(DataDROP, self).initialize(**kwargs)
        self._children = []

    # ===========================================================================
    # No data-related operations should actually be called in Container DROPs
    # ===========================================================================
    def getIO(self):
        return ErrorIO()

    @property
    def dataURL(self):
        raise OperationalError()

    def addChild(self, child):

        # Avoid circular dependencies between Containers
        if child == self.parent:
            raise InvalidRelationshipException(
                DROPRel(child, DROPLinkType.CHILD, self),
                "Circular dependency found",
            )

        logger.debug("Adding new child for %r: %r", self, child)

        self._children.append(child)
        child.parent = self

    def delete(self):
        # TODO: this needs more thinking. Probably a separate method to perform
        #       this recursive deletion will be needed, while this delete method
        #       will go hand-to-hand with the rest of the I/O methods above,
        #       which are currently raise a NotImplementedError
        if self._children:
            for c in [c for c in self._children if c.exists()]:
                c.delete()

    @property
    def expirationDate(self):
        if self._children:
            return heapq.nlargest(
                1, [c.expirationDate for c in self._children]
            )[0]
        return self._expirationDate

    @property
    def children(self):
        return self._children[:]

    def exists(self):
        if self._children:
            # TODO: Or should it be all()? Depends on what the exact contract of
            #       "exists" is
            return any([c.exists() for c in self._children])
        return True



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Utility functions for DROP I/O. This has been factored out from droputils
to avoid cyclic imports.
"""

import io
import logging
import pickle
import re
from typing import Any
import numpy as np

from dlg.data.io import OpenMode

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dlg.data.drops.data_base import DataDROP

logger = logging.getLogger(__name__)

# Used to check whether a command specifies via UID reference the path or
# data URL of an input or output
indexed_ipath_pattern = re.compile(r".*%i\[.+\].*")
indexed_opath_pattern = re.compile(r".*%o\[.+\].*")
indexed_idataurl_pattern = re.compile(r".*%iDataURL\[.+\].*")
indexed_odataurl_pattern = re.compile(r".*%oDataURL\[.+\].*")


def save_pickle(drop: "DataDROP", data: Any):
    """Saves a python object in pkl format"""
    pickle.dump(data, drop)


def load_pickle(drop: "DataDROP") -> Any:
    """Loads a pkl formatted data object stored in a DataDROP.
    Note: does not support streaming mode.
    """
    buf = io.BytesIO()
    desc = drop.open()
    while True:
        data = drop.read(desc)
        if not data:
            break
        buf.write(data)
    drop.close(desc)
    return pickle.loads(buf.getbuffer())


def save_npy(drop: "DataDROP", ndarray: np.ndarray, allow_pickle=False):
    """
    Saves a numpy ndarray to a drop in npy format
    """
    bio = io.BytesIO()
    # np.save accepts a "file-like" object which basically just requires
    # a .write() method. Try np.save(drop, array)
    np.save(bio, ndarray, allow_pickle=allow_pickle)
    dropio = drop.getIO()
    dropio.open(OpenMode.OPEN_WRITE)
    dropio.write(bio.getbuffer())
    dropio.close()


def save_numpy(drop: "DataDROP", ndarray: np.ndarray):
    save_npy(drop, ndarray)


def load_npy(drop: "DataDROP", allow_pickle=False) -> np.ndarray:
    """
    Loads a numpy ndarray from a drop in npy format
    """
    dropio = drop.getIO()
    dropio.open(OpenMode.OPEN_READ)
    res = np.load(io.BytesIO(dropio.buffer()), allow_pickle=allow_pickle)
    dropio.close()
    return res


def load_numpy(drop: "DataDROP", allow_pickle=True):
    return load_npy(drop, allow_pickle=allow_pickle)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Module contains shared memory manager which handles shared memory for a multi-threaded NodeManager.
"""
import logging

from dlg.shared_memory import DlgSharedMemory

LOGGER = logging.getLogger(__name__)


def _cleanup_block(session_id, name):
    mem = DlgSharedMemory(f"{session_id}_{name}")
    mem.close()
    mem.unlink()  # It is unlinking that is critical to freeing resources from the OS


def _close_block(session_id, name):
    mem = DlgSharedMemory(f"{session_id}_{name}")
    mem.close()


class DlgSharedMemoryManager:
    """
    Light class used by a NodeManager to log the existence of sharedmemory blocks.
    Unlinks all objects when requested
    """

    def __init__(self):
        self.drop_names = {}

    def register_session(self, name):
        """
        Registers a session with this memory manager
        """
        if str(name) in self.drop_names.keys():
            return
        self.drop_names[str(name)] = set()  # Handles duplicates

    def register_drop(self, name, session_id):
        """
        Adds a drop to the list of known shared memory blocks
        """
        if str(session_id) not in self.drop_names.keys():
            self.register_session(session_id)
        self.drop_names[str(session_id)].add(str(name))

    def shutdown_session(self, session_id):
        """
        Unlinks all memory blocks associated with a particular session.
        """
        if session_id in self.drop_names.keys():
            for drop in self.drop_names[session_id]:
                _cleanup_block(session_id, drop)

    def destroy_session(self, session_id):
        if session_id in self.drop_names.keys():
            for drop in self.drop_names[session_id]:
                _cleanup_block(session_id, drop)

    def shutdown_all(self):
        """
        Unlinks all shared memory blocks associated with this memory manager
        """
        for session_id in self.drop_names:
            self.destroy_session(session_id)
        self.drop_names = {}

    def __del__(self):
        if len(self.drop_names) > 0:
            self.shutdown_all()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import codecs
import json
import os
import sys
import time
import unittest

import pkg_resources
from dlg import droputils
from dlg import utils
from dlg.common import tool
from dlg.ddap_protocol import DROPStates
from dlg.manager.composite_manager import MasterManager
from dlg.manager.session import SessionStates
from dlg.testutils import ManagerStarter
from dlg.exceptions import NoSessionException
from test.manager import testutils

hostname = "localhost"

default_repro = {
    "rmode": "1",
    "RERUN": {
        "lg_blockhash": "x",
        "pgt_blockhash": "y",
        "pg_blockhash": "z",
    },
}
default_graph_repro = {
    "rmode": "1",
    "meta_data": {"repro_protocol": 0.1, "hashing_alg": "_sha3.sha3_256"},
    "merkleroot": "a",
    "RERUN": {
        "signature": "b",
    },
}


def add_test_reprodata(graph: list):
    for drop in graph:
        drop["reprodata"] = default_repro.copy()
    graph.append(default_graph_repro.copy())
    return graph


class DimAndNMStarter(ManagerStarter):
    def setUp(self):
        super(DimAndNMStarter, self).setUp()
        self.nm_info = self.start_nm_in_thread()
        self.dim_info = self.start_dim_in_thread()
        self.nm = self.nm_info.manager
        self.dim = self.dim_info.manager
        self.mm = MasterManager([hostname])

    def tearDown(self):
        self.mm.shutdown()
        self.dim_info.stop()
        self.nm_info.stop()


class TestMM(DimAndNMStarter, unittest.TestCase):
    def createSessionAndAddTypicalGraph(self, sessionId, sleepTime=0):
        graphSpec = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "island": hostname,
                "node": hostname,
                "consumers": ["B"],
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.SleepAndCopyApp",
                "sleep_time": sleepTime,
                "outputs": ["C"],
                "node": hostname,
                "island": hostname,
            },
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "island": hostname,
                "node": hostname,
            },
        ]
        add_test_reprodata(graphSpec)
        self.mm.createSession(sessionId)
        self.mm.addGraphSpec(sessionId, graphSpec)

    def test_createSession(self):
        sessionId = "lalo"
        self.mm.createSession(sessionId)
        self.assertEqual(1, len(self.nm.getSessionIds()))
        self.assertEqual(sessionId, self.nm.getSessionIds()[0])

    def test_addGraphSpec(self):
        sessionId = "lalo"

        # No node specified
        graphSpec = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        ]
        self.assertRaises(
            Exception, self.mm.addGraphSpec, sessionId, graphSpec
        )

        # Wrong node specified
        graphSpec = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": "unknown_host",
            }
        ]
        self.assertRaises(
            Exception, self.mm.addGraphSpec, sessionId, graphSpec
        )

        # No island specified
        graphSpec = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": hostname,
            }
        ]
        self.assertRaises(
            Exception, self.mm.addGraphSpec, sessionId, graphSpec
        )

        # Wrong island specified
        graphSpec = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": hostname,
                "island": "unknown_host",
            }
        ]
        self.assertRaises(
            Exception, self.mm.addGraphSpec, sessionId, graphSpec
        )

        # OK
        graphSpec = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": hostname,
                "island": hostname,
            }
        ]
        self.mm.createSession(sessionId)
        self.mm.addGraphSpec(sessionId, graphSpec)

        graphFromNM = self.nm.getGraph(sessionId)
        graphFromDIM = self.dim.getGraph(sessionId)
        graphFromMM = self.dim.getGraph(sessionId)
        self.assertDictEqual(graphFromNM, graphFromDIM)
        self.assertDictEqual(graphFromMM, graphFromDIM)

        self.assertEqual(1, len(graphFromMM))
        dropSpec = list(graphFromMM.values())[0]
        self.assertEqual("A", dropSpec["oid"])
        self.assertEqual("Data", dropSpec["categoryType"])
        self.assertEqual(
            "dlg.data.drops.memory.InMemoryDROP", dropSpec["dropclass"]
        )

    def test_deployGraph(self):
        sessionId = "lalo"
        self.createSessionAndAddTypicalGraph(sessionId)

        # Deploy now and get A and C
        self.mm.deploySession(sessionId)
        a, c = [self.nm._sessions[sessionId].drops[x] for x in ("A", "C")]

        data = os.urandom(10)
        with droputils.DROPWaiterCtx(self, c, 3):
            a.write(data)
            a.setCompleted()

        self.assertEqual(data, droputils.allDropContents(c))

    def test_deployGraphWithCompletedDOs(self):
        sessionId = "lalo"
        self.createSessionAndAddTypicalGraph(sessionId, sleepTime=2)

        # Deploy now and get A
        self.mm.deploySession(sessionId, completedDrops=["A"])
        c = self.nm._sessions[sessionId].drops["C"]

        # This should be happening before the sleepTime expires
        with droputils.DROPWaiterCtx(self, c, 5):
            pass

        self.assertEqual(DROPStates.COMPLETED, c.status)

    def test_sessionStatus(self):
        def assertSessionStatus(sessionId, status):
            sessionStatusMM = self.mm.getSessionStatus(sessionId)
            sessionStatusDIM = self.dim.getSessionStatus(sessionId)
            sessionStatusNM = self.nm.getSessionStatus(sessionId)
            self.assertEqual(1, len(sessionStatusMM))
            self.assertIn(hostname, sessionStatusMM)
            self.assertDictEqual(sessionStatusDIM, sessionStatusMM[hostname])
            self.assertEqual(
                sessionStatusNM, sessionStatusMM[hostname][hostname]
            )
            self.assertEqual(sessionStatusNM, status)

        sessionId = "lala"
        self.mm.createSession(sessionId)
        assertSessionStatus(sessionId, SessionStates.PRISTINE)

        sessionId = "lalo"
        self.createSessionAndAddTypicalGraph(sessionId)
        assertSessionStatus(sessionId, SessionStates.BUILDING)

        self.nm.deploySession(sessionId)
        assertSessionStatus(sessionId, SessionStates.RUNNING)

        a, c = [self.nm._sessions[sessionId].drops[x] for x in ("A", "C")]
        data = os.urandom(10)
        with droputils.DROPWaiterCtx(self, c, 3):
            a.write(data)
            a.setCompleted()

        assertSessionStatus(sessionId, SessionStates.FINISHED)

    def test_getGraph(self):
        sessionId = "lalo"
        self.createSessionAndAddTypicalGraph(sessionId)

        graphSpecFromMM = self.mm.getGraph(sessionId)
        self.assertEqual(3, len(graphSpecFromMM))
        for oid in ("A", "B", "C"):
            self.assertIn(oid, graphSpecFromMM)
        graphSepcFromNM = self.nm.getGraph(sessionId)
        graphSepcFromDIM = self.dim.getGraph(sessionId)
        self.assertDictEqual(graphSepcFromNM, graphSpecFromMM)
        self.assertDictEqual(graphSepcFromDIM, graphSpecFromMM)

    def test_getGraphStatus(self):
        def assertGraphStatus(sessionId, expectedStatus):
            graphStatusByDIM = self.dim.getGraphStatus(sessionId)
            graphStatusByDM = self.nm.getGraphStatus(sessionId)
            graphStatusByMM = self.mm.getGraphStatus(sessionId)
            self.assertDictEqual(graphStatusByDIM, graphStatusByMM)
            self.assertDictEqual(graphStatusByDIM, graphStatusByDM)
            for dropStatus in graphStatusByMM.values():
                self.assertEqual(expectedStatus, dropStatus["status"])

        sessionId = "lala"
        self.createSessionAndAddTypicalGraph(sessionId)
        self.mm.deploySession(sessionId)
        assertGraphStatus(sessionId, DROPStates.INITIALIZED)

        a, c = [self.nm._sessions[sessionId].drops[x] for x in ("A", "C")]
        data = os.urandom(10)
        with droputils.DROPWaiterCtx(self, c, 3):
            a.write(data)
            a.setCompleted()
        assertGraphStatus(sessionId, DROPStates.COMPLETED)


class TestREST(DimAndNMStarter, unittest.TestCase):
    def test_fullRound(self):
        """
        A test that exercises most of the REST interface exposed on top of the
        DataIslandManager
        """

        sessionId = "lala"
        restPort = 8989

        args = ["--port", str(restPort), "-N", hostname, "-qqq"]
        mmProcess = tool.start_process("mm", args)

        with testutils.terminating(mmProcess, 10):
            # Wait until the REST server becomes alive
            self.assertTrue(
                utils.portIsOpen("localhost", restPort, timeout=10),
                "REST server didn't come up in time",
            )

            # The DIM is still empty
            sessions = testutils.get(self, "/sessions", restPort)
            self.assertEqual(0, len(sessions))
            dimStatus = testutils.get(self, "", restPort)
            self.assertEqual(1, len(dimStatus["hosts"]))
            self.assertEqual(hostname, dimStatus["hosts"][0])
            self.assertEqual(0, len(dimStatus["sessionIds"]))

            # Create a session and check it exists
            testutils.post(
                self, "/sessions", restPort, '{"sessionId":"%s"}' % (sessionId)
            )
            sessions = testutils.get(self, "/sessions", restPort)
            self.assertEqual(1, len(sessions))
            self.assertEqual(sessionId, sessions[0]["sessionId"])
            self.assertDictEqual(
                {hostname: {hostname: SessionStates.PRISTINE}},
                sessions[0]["status"],
            )

            # Add this complex graph spec to the session
            # The UID of the two leaf nodes of this complex.js graph are T and S
            # Since the original complexGraph doesn't have node information
            # we need to add it manually before submitting -- otherwise it will
            # get rejected by the DIM.
            with pkg_resources.resource_stream(
                "test", "graphs/complex.js"
            ) as f:  # @UndefinedVariable
                complexGraphSpec = json.load(codecs.getreader("utf-8")(f))
            for dropSpec in complexGraphSpec:
                dropSpec["node"] = hostname
                dropSpec["island"] = hostname
            testutils.post(
                self,
                "/sessions/%s/graph/append" % (sessionId),
                restPort,
                json.dumps(complexGraphSpec),
            )
            self.assertEqual(
                {hostname: {hostname: SessionStates.BUILDING}},
                testutils.get(
                    self, "/sessions/%s/status" % (sessionId), restPort
                ),
            )

            # Now we deploy the graph...
            testutils.post(
                self,
                "/sessions/%s/deploy" % (sessionId),
                restPort,
                "completed=SL_A,SL_B,SL_C,SL_D,SL_K",
                mimeType="application/x-www-form-urlencoded",
            )
            self.assertEqual(
                {hostname: {hostname: SessionStates.RUNNING}},
                testutils.get(
                    self, "/sessions/%s/status" % (sessionId), restPort
                ),
            )

            # ...and write to all 5 root nodes that are listening in ports
            # starting at 1111
            msg = os.urandom(10)
            for i in range(5):
                utils.write_to(
                    "localhost", 1111 + i, msg, 2
                ), "Couldn't write data to localhost:%d" % (1111 + i)

            # Wait until the graph has finished its execution. We'll know
            # it finished by polling the status of the session
            while (
                SessionStates.RUNNING
                in testutils.get(
                    self, "/sessions/%s/status" % (sessionId), restPort
                )[hostname].values()
            ):
                time.sleep(0.2)

            self.assertEqual(
                {hostname: {hostname: SessionStates.FINISHED}},
                testutils.get(
                    self, "/sessions/%s/status" % (sessionId), restPort
                ),
            )
            testutils.delete(self, "/sessions/%s" % (sessionId), restPort)
            sessions = testutils.get(self, "/sessions", restPort)
            self.assertEqual(0, len(sessions))

            # Check log should not exist
            resp, _ = testutils._get("/sessions/not_exist/logs", 8000)
            self.assertEqual(resp.status, 404)

            # Check logs exist and there is content
            resp, _ = testutils._get("/sessions/{sessionId}/logs", restPort)
            self.assertGreater(int(resp.getheader("Content-Length")), 0)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import http.client
import json
import codecs

from dlg import utils


def _get(url, port):
    conn = http.client.HTTPConnection("localhost", port, timeout=3)
    conn.request("GET", "/api" + url)
    return conn.getresponse(), conn


def get(test, url, port):
    res, conn = _get(url, port)
    test.assertEqual(http.HTTPStatus.OK, res.status)
    jsonRes = json.load(codecs.getreader("utf-8")(res))
    res.close()
    conn.close()
    return jsonRes


def post(test, url, port, content=None, mimeType=None):
    conn = http.client.HTTPConnection("localhost", port, timeout=3)
    headers = {mimeType or "Content-Type": "application/json"} if content else {}
    conn.request("POST", "/api" + url, content, headers)
    res = conn.getresponse()
    test.assertEqual(http.HTTPStatus.OK, res.status)
    conn.close()


def delete(test, url, port):
    conn = http.client.HTTPConnection("localhost", port, timeout=3)
    conn.request("DELETE", "/api" + url)
    res = conn.getresponse()
    test.assertEqual(http.HTTPStatus.OK, res.status)
    conn.close()


class terminating(object):
    """
    A context manager that makes sure a process always exits.
    """

    def __init__(self, proc, timeout):
        self.proc = proc
        self.timeout = timeout

    def __enter__(self):
        return self.proc

    def __exit__(self, typ, val, traceback):
        utils.terminate_or_kill(self.proc, self.timeout)
        return False



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#


def __setupTestLogging():
    import os
    import logging

    # If we indicated a logging level for the tests we use that level
    # for our framework (and enable logging)
    level = os.environ.get("DALIUGE_TESTS_LOGLEVEL", None)
    if level:
        fmt = "%(asctime)-15s [%(levelname)5.5s] [%(threadName)15.15s] %(name)s#%(funcName)s:%(lineno)s %(message)s"
        logging.basicConfig(format=fmt, level=level)
    else:
        logging.root.addHandler(logging.NullHandler())


__setupTestLogging()
del __setupTestLogging



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Module implementing the PyFuncApp class"""

import ast
import base64
import collections
from enum import Enum
import importlib
import inspect
import json
import logging
import os
import pickle

from typing import Callable
import dill
from io import StringIO
from contextlib import redirect_stdout

from dlg import droputils, drop_loaders
from dlg.utils import serialize_data, deserialize_data
from dlg.named_port_utils import check_ports_dict, identify_named_ports
from dlg.apps.app_base import BarrierAppDROP
from dlg.exceptions import InvalidDropException
from dlg.meta import (
    dlg_string_param,
    dlg_enum_param,
    dlg_dict_param,
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
)

logger = logging.getLogger(__name__)


def serialize_func(f):
    if isinstance(f, str):
        parts = f.split(".")
        f = getattr(importlib.import_module(".".join(parts[:-1])), parts[-1])

    fser = dill.dumps(f)
    fdefaults = {"args": [], "kwargs": {}}
    adefaults = {"args": [], "kwargs": {}}
    a = inspect.getfullargspec(f)
    if a.defaults:
        fdefaults["kwargs"] = dict(
            zip(
                a.args[-len(a.defaults) :],
                [serialize_data(d) for d in a.defaults],
            )
        )
        adefaults["kwargs"] = dict(
            zip(a.args[-len(a.defaults) :], [d for d in a.defaults])
        )
    logger.debug(f"Introspection of function {f}: {a}")
    logger.debug("Defaults for function %r: %r", f, adefaults)
    return fser, fdefaults


def import_using_name(app, fname):
    logger.debug("Importing %s", fname)
    parts = fname.split(".")
    # If only one part check if builtin
    b = globals()["__builtins__"]
    logger.debug(f"Function: {parts[0]}: {hasattr(b, parts[0])}")
    if len(parts) < 2:
        logger.debug(f"Builtins: {type(b)}")
        logger.debug(f"Function {fname}: {hasattr(b, fname)}")
        if fname in b:
            return b[fname]
        else:
            msg = "%s is not builtin and does not contain a module name" % fname
            raise InvalidDropException(app, msg)
    elif parts[0] in b.keys():
        return b[parts[0]]
    else:
        if len(parts) > 1:
            if parts[-1] in ["__init__", "__class__"]:
                parts = parts[:-1]
            logger.debug("Recursive import: %s", parts)
            try:
                mod = importlib.import_module(parts[0], __name__)
            except ImportError as e:
                raise InvalidDropException(
                    app,
                    "Error when loading module %s: %s" % (parts[0], str(e)),
                )
            for m in parts[1:]:
                try:
                    logger.debug("Getting attribute %s", m)
                    mod = getattr(mod, m)
                except AttributeError as e:
                    try:
                        logger.debug(
                            "Trying to load backwards: %s",
                            ".".join(parts[:-1]),
                        )
                        mod = importlib.import_module(".".join(parts[:-1]), __name__)
                        mod = getattr(mod, parts[-1])
                        break
                    except ModuleNotFoundError:
                        # try again, sometimes fixes the namespace
                        mod = import_using_name(app, fname)
                        break
                    except Exception as e:
                        raise InvalidDropException(
                            app, "Problem importing module %s, %s" % (mod, e)
                        )
            logger.debug("Loaded module: %s", mod)
            return mod


def import_using_code(code):
    return dill.loads(code)


class DropParser(Enum):
    RAW = "raw"
    PICKLE = "pickle"
    EVAL = "eval"
    NPY = "npy"
    # JSON = "json"
    PATH = "path"  # input only
    DATAURL = "dataurl"  # input only


##
# @brief PythonMemberFunction
# @details A placeholder APP to aid construction of new class member function applications.
# This is mainly useful (and used) when starting a new workflow from scratch.
# @par EAGLE_START
# @param category PythonMemberFunction
# @param tag daliuge
# @param func_name object.__init__/String/ComponentParameter/NoPort/ReadWrite//False/False/Python function name
# @param func_code /String/ComponentParameter/NoPort/ReadWrite//False/False/Python function code, e.g. 'def function_name(args): return args'
# @param dropclass dlg.apps.pyfunc.PyFuncApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param object /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/object port
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/The allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @param input_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Input port parsing technique
# @param output_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Output port parsing technique
#     \~English Mapping from argname to default value. Should match only the last part of the argnames list.
#               Values are interpreted as Python code literals and that means string values need to be quoted.
# @par EAGLE_END
class PyMemberApp(BarrierAppDROP):
    """A placeholder member function that just aids the generation of the palette component"""

    pass


##
# @brief PyFuncApp
# @details An application that wraps a simple python function.
# The inputs of the application are treated as the arguments of the function.
# Conversely, the output of the function is treated as the output of the
# application. If the application has more than one output, the result of
# calling the function is treated as an iterable, with each individual object
# being written to its corresponding output.
# @par EAGLE_START
# @param category PythonApp
# @param tag template
# @param func_name /String/ComponentParameter/NoPort/ReadWrite//False/False/Python function name
# @param func_code /String/ComponentParameter/NoPort/ReadWrite//False/False/Python function code, e.g. 'def function_name(args): return args'
# @param dropclass dlg.apps.pyfunc.PyFuncApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/The allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @param input_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Input port parsing technique
# @param output_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Output port parsing technique
#     \~English Mapping from argname to default value. Should match only the last part of the argnames list.
#               Values are interpreted as Python code literals and that means string values need to be quoted.
# @par EAGLE_END
class PyFuncApp(BarrierAppDROP):
    """
    An application that wraps a simple python function.

    The inputs of the application are treated as the arguments of the function.
    Conversely, the output of the function is treated as the output of the
    application. If the application has more than one output, the result of
    calling the function is treated as an iterable, with each individual object
    being written to its corresponding output.

    Users indicate the function to be wrapped via the ``func_name`` parameter.
    In this case func_name needs to specify a function in the standard form

    ``module.function``

    and the module needs to be accessible on the PYTHONPATH of the DALiuGE
    engine. Note that the engine is expanding the standard PYTHONPATH with
    DLG_ROOT/code. That directory is always available, even if the engine is
    running in a docker container.

    Otherwise, users can also *send* over the python code using the ``func_code``
    parameter. The code needs to be base64-encoded and produced with the marshal
    module of the same Python version used to run DALiuGE.

    The positional onlyargs will be used in order of appearance.
    """

    component_meta = dlg_component(
        "PyFuncApp",
        "Py Func App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    func_name = dlg_string_param("func_name", None)
    # func_code = dlg_bytes_param("func_code", None) # bytes or base64 string
    input_parser: DropParser = dlg_enum_param(DropParser, "input_parser", DropParser.PICKLE)  # type: ignore
    output_parser: DropParser = dlg_enum_param(DropParser, "output_parser", DropParser.PICKLE)  # type: ignore
    func_arg_mapping = dlg_dict_param("func_arg_mapping", {})
    func_defaults = dlg_dict_param("func_defaults", {})
    f: Callable
    fdefaults: dict

    def _mixin_func_defaults(self):
        """
        func_defaults can be passed in through an argument, but this is only used for
        dlg_delayed and in combination with passing func_code. For dlg_delayed the function
        and its parameters might be computed on a different host than the where the delayed
        function is called and thus the function needs to be serialized.
        """
        logger.debug(f"Starting evaluation of func_defaults: {self.func_defaults}")
        if (
            isinstance(self.func_defaults, dict)
            and len(self.func_defaults) > 0
            and list(self.func_defaults.keys()) == ["kwargs", "args"]
        ):
            for arg in self.func_defaults["args"]:
                self.func_defaults["kwargs"][arg] = arg
                self.func_defaults = self.func_defaults["kwargs"]
        elif (
            isinstance(self.func_defaults, dict)
            and "kwargs" in self.func_defaults
            and isinstance(self.func_defaults["kwargs"], dict)
        ):
            self.func_defaults = self.func_defaults["kwargs"]
        # we came all this way, now assume that any resulting dict is correct
        if not isinstance(self.func_defaults, dict):
            logger.error(
                "Wrong format or type for function defaults for %s: %r, %r",
                self.f.__name__,
                self.func_defaults,
                type(self.func_defaults),
            )
            raise ValueError
        if self.input_parser is DropParser.PICKLE:
            # only values are pickled, get them unpickled
            for name, value in self.func_defaults.items():
                self.func_defaults[name] = deserialize_data(value)
        # the fn_defaults are used afterwards, we'll drop the func_defaults
        logger.debug("fn_defaults %s", self.fn_defaults)
        logger.debug("func_defaults %s", self.func_defaults)
        self.fn_defaults = self.func_defaults

    def _init_fn_defaults(self):
        """
        Inititalize self.fn_defaults dictionary from values provided.
        Multiple options exist and some are here for compatibility.
        """
        logger.debug(f"Starting evaluation of function signature")
        self.argsig = inspect.signature(self.f)
        self.argnames = list(self.argsig.parameters.keys())
        logger.debug("Function signature: %s", self.argsig)
        args = list(self.argsig.parameters.keys())
        self.fn_nargs = len(self.argsig.parameters)
        # set defaults
        self.fn_nposkw = 0
        self.poskw = {}
        self.fn_npos = 0
        self.posonly = {}
        self.kwonly = {}
        self.fn_nkw = 0
        self.varargs = False
        self.varkw = False
        self.fn_ndef = 0
        for k, p in self.argsig.parameters.items():
            if not isinstance(p, str):
                if p.kind == p.POSITIONAL_OR_KEYWORD:
                    self.fn_nposkw += 1
                    self.poskw.update({k: p})
                elif p.kind == p.POSITIONAL_ONLY:
                    self.fn_npos += 1
                    self.posonly.update({k: p})
                elif p.kind == p.KEYWORD_ONLY:
                    self.fn_nkw += 1
                    self.kwonly.update({k: p})
                elif p.kind == p.VAR_POSITIONAL:
                    self.varargs = True
                elif p.kind == p.VAR_KEYWORD:
                    self.varkw = True
                if p.default != inspect._empty:
                    self.arguments_defaults.append(p.default)
                    self.fn_ndef += 1
                else:
                    self.arguments_defaults.append(None)
                    self.fn_ndef += 1

        logger.debug("Got signature for function %s %s", self.f, self.argsig)
        logger.debug("Got default values for arguments %s", self.arguments_defaults)
        self.fn_defaults = self.arguments_defaults
        logger.debug(f"initialized fn_defaults with {self.fn_defaults}")

    def _clean_applicationArgs(self):
        """
        remove function definition arguments in applicationArgs
        """
        self.func_def_keywords = [
            "func_code",
            "func_name",
            "func_arg_mapping",
            "input_parser",
            "output_parser",
            "func_defaults",
            "pickle",
        ]

        for kw in self.func_def_keywords:
            if kw in self._applicationArgs:  # these are the preferred ones now
                if isinstance(
                    self._applicationArgs[kw]["value"],
                    bool
                    or self._applicationArgs[kw]["value"]
                    or self._applicationArgs[kw]["precious"],
                ):
                    # only transfer if there is a value or precious is True
                    self._applicationArgs.pop(kw)

    def initialize_with_func_code(self):
        """
        This function takes over if code is passed in through an argument.
        """
        if not isinstance(self.func_code, bytes):
            self.func_code = base64.b64decode(self.func_code.encode("utf8"))
        self.f = import_using_code(self.func_code)
        self._init_fn_defaults()
        # make sure defaults are dicts
        self._mixin_func_defaults()
        if isinstance(self.func_arg_mapping, str):
            self.func_arg_mapping = ast.literal_eval(self.func_arg_mapping)

    def initialize(self, **kwargs):
        """
        The initialization of a function component is mainly dealing with mapping
        inputs and provided applicationArgs to the function arguments. All of this
        should be driven by matching names.
        """
        BarrierAppDROP.initialize(self, **kwargs)

        env = os.environ.copy()
        env.update({"DLG_UID": self._uid})
        if self._dlg_session_id:
            env.update({"DLG_SESSION_ID": self._dlg_session_id})
        elif "dlg_session_id" in kwargs:
            self._dlg_session_id = kwargs["dlg_session_id"]
            env.update({"DLG_SESSION_ID": self._dlg_session_id})

        self._applicationArgs = self._popArg(kwargs, "applicationArgs", {})

        self.func_code = self._popArg(kwargs, "func_code", None)
        self.arguments_defaults = []

        # backwards compatibility
        if "pickle" in self._applicationArgs:
            if self._applicationArgs["pickle"]["value"] == True:
                self.input_parser = DropParser.PICKLE
                self.output_parser = DropParser.PICKLE
            else:
                self.input_parser = DropParser.EVAL
                self.output_parser = DropParser.EVAL
            self._applicationArgs.pop("pickle")

        self._clean_applicationArgs()

        self.num_args = len(
            self._applicationArgs
        )  # number of additional arguments provided

        if not self.func_name and not self.func_code:
            raise InvalidDropException(
                self, "No function specified (either via name or code)"
            )

        # Lookup function or import bytecode as a function
        if not self.func_code:
            self.f = import_using_name(self, self.func_name)
            self._init_fn_defaults()
        else:
            self.initialize_with_func_code()

        logger.info(f"Args summary for '{self.func_name}':")
        logger.info(f"Args: {self.argnames}")
        logger.info(f"Args defaults:  {self.fn_defaults}")
        logger.info(f"Args pos/kw: {list(self.poskw.keys())}")
        logger.info(f"Args keyword only: {list(self.kwonly.keys())}")
        logger.info(f"VarArgs allowed:  {self.varargs}")
        logger.info(f"VarKwds allowed:  {self.varkw}")

        # Mapping between argument name and input drop uids
        logger.debug(f"Input mapping provided: {self.func_arg_mapping}")
        self._recompute_data = {}

    def run(self):
        """
        Function positional and keyword argument treatment:

        Function arguments can be provided in three different ways:
        1) Through an input port
        2) By specifying ApplicationArgs (one for each argument)
        3) Through defaults at the time of function definition

        The priority follows the list above with input ports overruling the others.
        Function arguments in Python can be passed as positional, kw-value, positional
        only, kw-value only, and catch-all args and kwargs, which don't provide any
        hint about the names of accepted parameters. All of them are now supported. If
        positional arguments or kw-value arguments are provided by the user, but are
        not explicitely defined in the function signiture AND args and/or kwargs are
        allowed then these arguments are passed to the function. For args this is
        somewhat risky, since the order is relevant and in this code derived from the
        order defined in the graph (same order as defined in the component description).

        Input ports will NOT be used by index (anymore), but by using the name of the port.
        Since each input port requires an associated data drop, this provides a unique
        mapping. This also allows to pass values to any function argument through a port.

        """
        funcargs = {}
        # Inputs are un-pickled and treated as the arguments of the function
        # Their order must be preserved, so we use an OrderedDict
        if self.input_parser is DropParser.PICKLE:
            # all_contents = lambda x: pickle.loads(droputils.allDropContents(x))
            all_contents = drop_loaders.load_pickle
        elif self.input_parser is DropParser.EVAL:

            def optionalEval(x):
                # Null and Empty Drops will return an empty byte string
                # which should propogate back to None
                content: str = droputils.allDropContents(x).decode("utf-8")
                return ast.literal_eval(content) if len(content) > 0 else None

            all_contents = optionalEval
        elif self.input_parser is DropParser.NPY:
            all_contents = drop_loaders.load_npy
        elif self.input_parser is DropParser.PATH:
            all_contents = lambda x: x.path
        elif self.input_parser is DropParser.DATAURL:
            all_contents = lambda x: x.dataurl
        else:
            raise ValueError(self.input_parser.__repr__())

        inputs = collections.OrderedDict()
        for uid, drop in self._inputs.items():
            inputs[uid] = all_contents(drop)

        outputs = collections.OrderedDict()
        for uid, drop in self._outputs.items():
            if self.output_parser is DropParser.PATH:
                outputs[uid] = drop.path
            else:
                outputs[uid] = None

        # Keyword arguments are made up of the default values plus the inputs
        # that match one of the keyword argument names
        # if defaults dict has not been specified at all we'll go ahead anyway

        # 1. Fill arguments with rest of inputs
        logger.debug(f"available inputs: {inputs}")

        posargs = list(self.posonly.keys()) + list(self.poskw.keys())
        kwargs = {}
        pargs = []
        # fill the pargsDict with positional and poskw arguments and defaults
        pargsDict = {k: v.default for k, v in self.posonly.items()}
        pargsDict.update({k: v.default for k, v in self.poskw.items()})
        logger.debug("Initial pos_kwargs dictionary: %s", pargsDict)
        # fill the keyargsDict with kwonly arguments and defaults
        keyargsDict = {k: v.default for k, v in self.kwonly.items()}
        logger.debug("Initial kwonly dictionary: %s", self.kwonly)

        # replace default argument values with provided values
        if "applicationArgs" in self.parameters:
            appArgs = self.parameters[
                "applicationArgs"
            ]  # we'll pop the identified ones
            _dum = [appArgs.pop(k) for k in self.func_def_keywords if k in appArgs]
            logger.debug(
                "Default keyword arguments removed: %s",
                [i for i in _dum],
            )
            # update the positional args
            pargsDict.update(
                {k: self.parameters[k] for k in pargsDict if k in self.parameters}
            )
            # if defined in both we use AppArgs values
            for k in appArgs:
                # check value type and interpret
                if appArgs[k]["type"] in ["Json", "Complex"]:
                    try:
                        value = ast.literal_eval(appArgs[k]["value"])
                        logger.debug(
                            f"Evaluated %s to %s",
                            appArgs[k]["value"],
                            type(value),
                        )
                        appArgs[k]["value"] = value
                    except ValueError:
                        logger.error("Unable to evaluate %s", appArgs[k]["value"])
                else:
                    value = appArgs[k]["value"]
                if k in pargsDict:
                    pargsDict.update({k: value})

            _ = [appArgs.pop(k) for k in pargsDict if k in appArgs]
            logger.debug("Updated posargs dictionary: %s", pargsDict)

            keyargsDict.update(
                {k: appArgs[k]["value"] for k in keyargsDict if k in appArgs}
            )
            logger.debug("Updated keyargs dictionary: %s", keyargsDict)

            # 2. put all remaining arguments into *args and **kwargs
            # TODO: This should only be done if the function signature allows it
            vparg = []
            vkarg = {}
            logger.debug(f"Remaining AppArguments {appArgs}")
            for arg in appArgs:
                if appArgs[arg]["type"] in ["Json", "Complex"]:
                    value = ast.literal_eval(appArgs[arg]["value"])
                else:
                    value = appArgs[arg]["value"]
                if appArgs[arg]["positional"]:
                    vparg.append(value)
                else:
                    vkarg.update({arg: value})

            # TODO: check where this is defined in signature
            self.arguments = inspect.getfullargspec(self.f)
            if self.arguments.varargs:
                logger.debug("Adding remaining *args to pargs %s", vparg)
                pargs.extend(vparg)
            if self.arguments.varkw:
                logger.debug("Adding remaining **kwargs to funcargs: %s", vkarg)
                funcargs.update(vkarg)

        # 3. replace default argument values with named input ports
        # TODO: investigate performing inputs and outputs in a single call
        if "inputs" in self.parameters and check_ports_dict(self.parameters["inputs"]):
            check_len = min(
                len(inputs),
                self.fn_nargs + self.fn_nkw,
            )
            inputs_dict = collections.OrderedDict()
            for inport in self.parameters["inputs"]:
                key = list(inport.keys())[0]
                inputs_dict[key] = {"name": inport[key], "path": inputs[key]}
            kwargs.update(
                identify_named_ports(
                    inputs_dict,
                    posargs,
                    pargsDict,
                    keyargsDict,
                    check_len=check_len,
                    mode="inputs",
                )
            )
        else:
            # Just as a fallback using the index, but this is risky!
            for i in range(min(len(inputs), self.fn_nargs)):
                kwargs.update({self.argnames[i]: list(inputs.values())[i]})

        # 4. replace default argument values with named output ports
        if "outputs" in self.parameters and check_ports_dict(
            self.parameters["outputs"]
        ):
            check_len = min(len(outputs), self.fn_nargs + self.fn_nkw)
            outputs_dict = collections.OrderedDict()
            for outport in self.parameters["outputs"]:
                key = list(outport.keys())[0]
                outputs_dict[key] = {
                    "name": outport[key],
                    "path": outputs[key],
                }

            kwargs.update(
                identify_named_ports(
                    outputs_dict,
                    posargs,
                    pargsDict,
                    keyargsDict,
                    check_len=check_len,
                    mode="outputs",
                )
            )
        logger.debug(f"Updating funcargs with values from pargsDict {pargsDict}")
        funcargs.update(pargsDict)

        logger.debug(f"Updating funcargs with values from named ports {kwargs}")
        funcargs.update(kwargs)

        self._recompute_data["args"] = funcargs.copy()

        # 5. remove self argument if this is the initializer.
        if (
            self.func_name is not None
            and self.func_name.split(".")[-1] in ["__init__", "__class__"]
            and "self" in funcargs
        ):
            funcargs.pop("self")
        logger.info(f"Running {self.func_name} with *{pargs} **{funcargs}")

        # 6. prepare for execution
        # we capture and log whatever is produced on STDOUT
        capture = StringIO()
        try:
            bind = self.argsig.bind(*pargs, **funcargs)
            logger.debug("Bound arguments: %s", bind)
        except TypeError as e:
            logger.error("Binding of arguments failed: %s", e)
            raise

        # Here is where the function is actually executed
        with redirect_stdout(capture):
            result = self.f(*bind.args, **bind.kwargs)

        logger.debug("Returned result from %s: %s", self.func_name, result)
        logger.info(
            f"Captured output from function app '{self.func_name}': {capture.getvalue()}"
        )
        logger.debug(f"Finished execution of {self.func_name}.")

        # Depending on how many outputs we have we treat our result
        # as an iterable or as a single object. Each result is pickled
        # and written to its corresponding output
        self.write_results(result)

    def write_results(self, result):
        outputs = self.outputs
        if len(outputs) > 0:
            if len(outputs) == 1:
                result = [result]
            for r, o in zip(result, outputs):
                if self.output_parser is DropParser.PICKLE:
                    logger.debug(f"Writing pickeled result {type(r)} to {o}")
                    o.write(pickle.dumps(r))
                elif self.output_parser is DropParser.EVAL:
                    o.write(repr(r).encode("utf-8"))
                elif self.output_parser is DropParser.NPY:
                    drop_loaders.save_npy(o, r)
                else:
                    ValueError(self.output_parser.__repr__())

    def generate_recompute_data(self):
        for name, val in self._recompute_data.items():
            try:
                json.dumps(val)
            except TypeError as e:
                logger.debug(e)
                self._recompute_data[name] = repr(val)
        return self._recompute_data



from dlg.drop import track_current_drop
from dlg.apps.app_base import BarrierAppDROP
from dlg.exceptions import InvalidDropException


##
# @brief Branch
# @details A conditional branch to control flow
# @par EAGLE_START
# @param category Branch
# @param tag template
# @param dropclass dlg.apps.simple.SimpleBranch/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/the allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @param dummy_input /Object/ApplicationArgument/InputPort/ReadWrite//False/False/Dummy input port
# @param dummy0 /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/Dummy output port
# @param dummy1 /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/Dummy output port
# @param input_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Input port parsing technique
# @param output_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Output port parsing technique
# @par EAGLE_END
class BranchAppDrop(BarrierAppDROP):
    """
    A special kind of application with exactly two outputs. After normal
    execution, the application decides whether a certain condition is met.
    If the condition is met, the first output is considered as COMPLETED,
    while the other is moved to SKIPPED state, and vice-versa.
    """

    @track_current_drop
    def execute(self, _send_notifications=True):
        if len(self._outputs) != 2:
            raise InvalidDropException(
                self,
                f"BranchAppDrops should have exactly 2 outputs, not {len(self._outputs)}",
            )
        BarrierAppDROP.execute(self, _send_notifications=False)
        self.outputs[1 if self.condition() else 0].skip()
        self._notifyAppIsFinished()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import tempfile
import threading
import unittest

from dlg import exceptions
from dlg.common.deployment_methods import DeploymentMethods
from dlg.exceptions import InvalidGraphException

from dlg.manager import constants
from dlg.manager.client import NodeManagerClient, DataIslandManagerClient
from dlg.manager.composite_manager import DataIslandManager
from dlg.manager.node_manager import NodeManager
from dlg.manager.rest import NMRestServer, CompositeManagerRestServer
from dlg.restutils import RestClient

default_repro = {
    "rmode": "1",
    "RERUN": {
        "lg_blockhash": "x",
        "pgt_blockhash": "y",
        "pg_blockhash": "z",
    },
}
default_graph_repro = {
    "rmode": "1",
    "meta_data": {"repro_protocol": 0.1, "hashing_alg": "_sha3.sha3_256"},
    "merkleroot": "a",
    "RERUN": {
        "signature": "b",
    },
}


def add_test_reprodata(graph: list):
    for drop in graph:
        drop["reprodata"] = default_repro.copy()
    graph.append(default_graph_repro.copy())
    return graph


hostname = "localhost"


class TestRest(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.dm = NodeManager(False)
        self._dm_server = NMRestServer(self.dm)
        self._dm_t = threading.Thread(
            target=self._dm_server.start,
            args=(hostname, constants.NODE_DEFAULT_REST_PORT),
        )
        self._dm_t.start()

        self.dim = DataIslandManager(dmHosts=[hostname])
        self._dim_server = CompositeManagerRestServer(self.dim)
        self._dim_t = threading.Thread(
            target=self._dim_server.start,
            args=(hostname, constants.ISLAND_DEFAULT_REST_PORT),
        )
        self._dim_t.start()

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self._dm_server.stop()
        self._dm_t.join()
        self.dm.shutdown()
        self.assertFalse(self._dm_t.is_alive())

        self._dim_server.stop()
        self._dim_t.join()
        self.dim.shutdown()
        self.assertFalse(self._dim_t.is_alive())

    def test_index(self):
        # Just check that the HTML pages load properly
        with RestClient(
            hostname, constants.NODE_DEFAULT_REST_PORT, timeout=10
        ) as c:
            c._GET("/")
            c._GET("/session")

    def test_errtype(self):
        sid = "lala"
        c = NodeManagerClient(hostname)
        c.createSession(sid)
        gempty = [{}]
        add_test_reprodata(gempty)
        # already exists
        self.assertRaises(
            exceptions.SessionAlreadyExistsException, c.createSession, sid
        )

        # different session
        self.assertRaises(
            exceptions.NoSessionException, c.addGraphSpec, sid + "x", [{}]
        )

        # invalid dropspec, it has no oid/type (is completely empty actually)
        self.assertRaises(
            exceptions.InvalidGraphException, c.addGraphSpec, sid, gempty
        )

        # invalid dropspec, app doesn't exist
        self.assertRaises(
            exceptions.InvalidGraphException,
            c.addGraphSpec,
            sid,
            [
                {
                    "oid": "a",
                    "categoryType": "Application",
                    "dropclass": "doesnt.exist",
                    "reprodata": default_repro.copy(),
                },
                default_graph_repro.copy(),
            ],
        )

        # invalid state, the graph status is only queried when the session is running
        self.assertRaises(
            exceptions.InvalidSessionState, c.getGraphStatus, sid
        )

        # valid dropspec, but the socket listener app doesn't allow inputs
        c.addGraphSpec(
            sid,
            [
                {
                    "categoryType": "socket",
                    "oid": "a",
                    "inputs": ["b"],
                    "reprodata": default_repro.copy(),
                },
                {
                    "oid": "b",
                    "categoryType": "Data",
                    "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                    "reprodata": default_repro.copy(),
                },
                default_graph_repro.copy(),
            ],
        )
        self.assertRaises(
            exceptions.InvalidRelationshipException, c.deploySession, sid
        )

        # And here we point to an unexisting file, making an invalid drop
        c.destroySession(sid)
        c.createSession(sid)
        fname = tempfile.mktemp()
        c.addGraphSpec(
            sid,
            [
                {
                    "categoryType": "Data",
                    "dropclass": "dlg.data.drops.file.FileDROP",
                    "oid": "a",
                    "filepath": fname,
                    "check_filepath_exists": True,
                    "reprodata": default_repro.copy(),
                },
                default_graph_repro.copy(),
            ],
        )
        self.assertRaises(
            exceptions.InvalidDropException, c.deploySession, sid
        )

    def test_recursive(self):
        sid = "lala"
        c = DataIslandManagerClient(hostname)
        c.createSession(sid)

        # invalid dropspec, app doesn't exist
        # This is not checked at the DIM level but only at the NM level
        # The exception should still pass through though
        with self.assertRaises(exceptions.SubManagerException) as cm:
            c.addGraphSpec(
                sid,
                [
                    {
                        "oid": "a",
                        "categoryType": "Application",
                        "dropclass": "doesnt.exist",
                        "node": hostname,
                        "reprodata": default_repro.copy(),
                    },
                    default_graph_repro.copy(),
                ],
            )
        ex = cm.exception
        self.assertTrue(hostname in ex.args[0])
        self.assertTrue(
            isinstance(ex.args[0][hostname], InvalidGraphException)
        )

    def test_reprodata_get(self):
        """
        Tests deploying an incredibly basic graph with and without reprodata
        Then querying the manager for that reprodata.
        """
        sid = "1234"
        c = NodeManagerClient(hostname)
        graph_spec = [
            {
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "oid": "a",
                "reprodata": default_repro.copy(),
            },
            default_graph_repro.copy(),
        ]
        # Test with reprodata
        c.createSession(sid)
        c.addGraphSpec(sid, graph_spec)
        c.deploySession(sid, completed_uids=["a"])
        response = c.session_repro_data(sid)
        self.assertIsNotNone(
            response["graph"]["a"]["reprodata"]["RERUN"]["rg_blockhash"]
        )
        self.assertIsNotNone(response["reprodata"])
        c.destroySession(sid)
        # Test without reprodata
        graph_spec = graph_spec[0:1]
        graph_spec[0].pop("reprodata")
        c.createSession(sid)
        c.addGraphSpec(sid, graph_spec)
        c.deploySession(sid, completed_uids=["a"])
        response = c.session_repro_data(sid)
        self.assertEqual(
            {
                "a": {
                    "oid": "a",
                    "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                    "categoryType": "Data",
                }
            },
            response["graph"],
        )
        self.assertEqual({}, response["reprodata"])

    def test_reprostatus_get(self):
        # Test with reprodata
        sid = "1234"
        c = NodeManagerClient(hostname)
        graph_spec = [
            {
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "oid": "a",
                "reprodata": default_repro.copy(),
            },
            default_graph_repro.copy(),
        ]
        c.createSession(sid)
        c.addGraphSpec(sid, graph_spec)
        c.deploySession(sid, completed_uids=["a"])
        response = c.session_repro_status(sid)
        self.assertTrue(response)
        c.destroySession(sid)
        # Test without reprodata
        graph_spec = graph_spec[0:1]
        graph_spec[0].pop("reprodata")
        c.createSession(sid)
        c.addGraphSpec(sid, graph_spec)
        c.deploySession(sid, completed_uids=["a"])
        response = c.session_repro_status(sid)
        self.assertTrue(response)
        c.destroySession(sid)

    def test_submit_method(self):
        c = NodeManagerClient(hostname)
        response = c.get_submission_method()
        self.assertEqual({"methods": [DeploymentMethods.BROWSER, DeploymentMethods.SERVER]}, response)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2019
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module tests the helm chart translation and deployment functionality.
"""
import json
import os
import sys
import tempfile
import unittest

import yaml
from dlg.common.k8s_utils import check_k8s_env

if check_k8s_env():
    from dlg.common.version import version as dlg_version
    from dlg.deploy.helm_client import HelmClient


@unittest.skipIf(
    sys.version_info <= (3, 8), "Copying temp files fail on Python < 3.7"
)
@unittest.skipIf(not check_k8s_env(), "K8s is not available")
class TestHelmClient(unittest.TestCase):
    def test_create_default_helm_chart(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            helm_client = HelmClient(
                deploy_dir=tmp_dir, deploy_name="my_fun_name"
            )
            helm_client.create_helm_chart("[]")
            chart_file_name = os.path.join(
                helm_client._chart_dir, "Chart.yaml"
            )
            with open(chart_file_name, "r", encoding="utf-8") as chart_file:
                chart_data = yaml.safe_load(chart_file)
                self.assertEqual(helm_client._chart_name, chart_data["name"])
                self.assertEqual(dlg_version, chart_data["appVersion"])

    @unittest.skip
    def test_custom_ports(self):
        pass

    def test_create_single_node_helm_chart(self):
        pg = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.SleepApp",
                "inputs": ["A"],
                "outputs": ["C"],
            },
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
        ]
        for drop in pg:
            drop["node"] = "localhost"
            drop["island"] = "localhost"
        with tempfile.TemporaryDirectory() as tmp_dir:
            helm_client = HelmClient(
                deploy_dir=tmp_dir, deploy_name="dlg-test"
            )
            helm_client.create_helm_chart(json.dumps(pg), co_host=False)
            self.assertEqual(pg, json.loads(helm_client._physical_graph_file))
            self.assertEqual(1, helm_client._num_machines)

    def test_create_multi_node_helm_chart(self):
        pg = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": "localhost",
                "island": "localhost",
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.SleepApp",
                "inputs": ["A"],
                "outputs": ["C"],
                "node": "localhost",
                "island": "localhost",
            },
            {
                "oid": "D",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.SleepApp",
                "inputs": ["A"],
                "outputs": ["E"],
                "node": "127.0.0.2",
                "island": "127.0.0.2",
            },
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": "localhost",
                "island": "localhost",
            },
            {
                "oid": "E",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "node": "127.0.0.2",
                "island": "127.0.0.2",
            },
        ]
        with tempfile.TemporaryDirectory() as tmp_dir:
            helm_client = HelmClient(
                deploy_dir=tmp_dir, deploy_name="dlg_test"
            )
            helm_client.create_helm_chart(json.dumps(pg), co_host=False)
            # TODO: Assert translation works
            self.assertEqual(2, helm_client._num_machines)

    @unittest.skip
    def test_submit_job(self):
        self.fail("Test not yet implemented")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os
import sys
import uuid
import threading

from dlg.data.drops.memory import InMemoryDROP
from dlg.data.drops.directorycontainer import DirectoryContainer

from .chilesdo import Split, Clean, SourceFlux

LOCAL_FILES = os.path.dirname(os.path.realpath(__file__))
CASAPY = "/home/jenkins/casa-release-4.4.0-el6/"
VIS_ROOT = "/mnt/chiles-imaging/DataFiles/"
VIS_OUT = "/mnt/chiles-output/vis/"
CUBE_OUT = "/mnt/chiles-output/cube/"
CUBE_NAME = "cube1408~1412"

VIS = [
    (
        VIS_ROOT
        + "20131025_951_4_FINAL_PRODUCTS/20131025_951_4_calibrated_deepfield.ms",
        VIS_OUT + "20131025_951_4/",
    ),
    (
        VIS_ROOT
        + "20131031_951_4_FINAL_PRODUCTS/20131031_951_4_calibrated_deepfield.ms",
        VIS_OUT + "20131031_951_4/",
    ),
    (
        VIS_ROOT
        + "20131121_946_6_FINAL_PRODUCTS/20131121_946_6_calibrated_deepfield.ms",
        VIS_OUT + "20131121_946_6/",
    ),
    (
        VIS_ROOT
        + "20140105_946_6_FINAL_PRODUCTS/20140105_946_6_calibrated_deepfield.ms",
        VIS_OUT + "20140105_946_6/",
    ),
]


class Barrier(object):
    def __init__(self, drop):
        self._evt = threading.Event()
        drop.addConsumer(self)

    def dropCompleted(self, drop, state):
        self._evt.set()

    def wait(self, timeout=None):
        return self._evt.wait(timeout)


if __name__ == "__main__":
    try:

        vis_in_a = []

        flux_out = InMemoryDROP(uuid.uuid1(), uuid.uuid1())

        flux = SourceFlux(uuid.uuid1(), uuid.uuid1(), casapy_path=CASAPY)

        cl = Clean(
            uuid.uuid1(),
            uuid.uuid1(),
            field="deepfield",
            mode="frequency",
            restfreq="1420.405752MHz",
            nchan=-1,
            start="",
            width="",
            interpolation="nearest",
            gain=0.1,
            imsize=[256],
            cell=["1.0arcsec"],
            phasecenter="10h01m53.9,+02d24m52s",
            weighting="natural",
            casapy_path=CASAPY,
        )

        image_out = DirectoryContainer(
            uuid.uuid1(), uuid.uuid1(), dirname=CUBE_OUT + CUBE_NAME, check_exists=False
        )
        cl.addOutput(image_out)
        flux.addInput(image_out)
        flux.addOutput(flux_out)

        for v in VIS:
            vis_in = DirectoryContainer(uuid.uuid1(), uuid.uuid1(), dirname=v[0])
            split_out = DirectoryContainer(
                uuid.uuid1(), uuid.uuid1(), dirname=v[1], check_exists=False
            )

            vis_in_a.append(vis_in)

            sp = Split(
                uuid.uuid1(),
                uuid.uuid1(),
                regridms=True,
                restfreq="1420.405752MHz",
                mode="frequency",
                nchan=256,
                outframe="lsrk",
                interpolation="linear",
                start="1408 MHz",
                width="1412 kHz",
                copy=False,
                casapy_path=CASAPY,
            )

            sp.addInput(vis_in)
            sp.addOutput(split_out)

            cl.addInput(split_out)

        # start
        for i in vis_in_a:
            i.setCompleted()

        # wait for flux value to be calculated
        b = Barrier(flux_out)
        res = b.wait(6000)
        if res == False:
            raise Exception("imaging timeout!")

    except Exception as e:
        import traceback

        traceback.print_exc()
        sys.exit(-1)

    sys.exit(0)
